from django.core.management.base import BaseCommand
from django.db import transaction
from django.db.models import Count, Max, F, OuterRef, Subquery, ExpressionWrapper, IntegerField
from django.db.models.functions import Coalesce

from src.management.commands.default import delete_graphs
from src.management.commands.core import create_arc
from src.models import Graph, Vertex, Arc
import time


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('min', type=int)
        parser.add_argument('max', type=int)

    def handle(self, *args, **options):
        delete_graphs("all")

        min = options['min']
        max = options['max']

        s_time = time.time()
        for n in range(min, max + 1):
            start_time = time.time()
            self.create_path(n)
            end_time = time.time()
            elapsed_time = end_time - start_time

            print("Čas trvání", n, "vrcholů:", elapsed_time, "sekund")
        e_time = time.time()
        eed_time = e_time - s_time

        print("Celkový čas trvání:", eed_time, "sekund")

    def create_path(self, n):
        with transaction.atomic():
            for i in range(2 ** (n-1)):
                binary_number = bin(i)[2:].zfill(n)
                graph = self.create_graph(n, binary_number)

                arc_count_in_subquery = (
                    Arc.objects
                    .filter(vertex_in=OuterRef('pk'))
                    .values('vertex_in')
                    .annotate(arc_count_in=Count('label'))
                    .values('arc_count_in')
                    .order_by('-arc_count_in')
                    [:1]
                )

                arc_count_out_subquery = (
                    Arc.objects
                    .filter(vertex_out=OuterRef('pk'))
                    .values('vertex_out')
                    .annotate(arc_count_out=Count('label'))
                    .values('arc_count_out')
                    .order_by('-arc_count_out')
                    [:1]
                )

                # Získání maximálního počtu arků pro každý vrchol
                vertices = (
                    Vertex.objects
                    .filter(graph=graph)
                    .annotate(arc_count_in=Coalesce(Subquery(arc_count_in_subquery, output_field=IntegerField()), 0))
                    .annotate(arc_count_out=Coalesce(Subquery(arc_count_out_subquery, output_field=IntegerField()), 0))
                    .annotate(max_arc_count=ExpressionWrapper(
                        Max(F('arc_count_in'), F('arc_count_out')),
                        output_field=IntegerField()
                    ))
                    .order_by('max_arc_count', 'label')
                )
                for num, v in enumerate(vertices):
                    v.label = num + 1
                    v.save()
                graph.update()

                if i >= 2 ** (n - 1) / 2 - 1:
                    break

    @staticmethod
    def create_graph(n, binary_number):
        prev_vertex = None
        arc_num = 0
        graph = Graph.objects.create(
            graph_type=Graph.PATH,
            binary=binary_number[1:],
        )

        for j, bit in enumerate(binary_number):
            vertex = Vertex.objects.create(graph=graph, index=j + 1)
            if prev_vertex:
                arc_num = create_arc(
                    vertex_in=prev_vertex if bit == '0' else vertex,
                    vertex_out=vertex if bit == '0' else prev_vertex,
                    arc_num=arc_num,
                    label=n + arc_num
                )
            prev_vertex = vertex
        return graph


        # for i in range(2 ** (n - 1)):
        #     binary_number = bin(i)[2:].zfill(n)
        #
        #     if i >= 2 ** (n - 1) / 2 - 1:
        #         break