from django.db import transaction
from django.db.models import OuterRef, Count, Subquery, ExpressionWrapper, IntegerField, F, Max
from django.db.models.functions import Coalesce

from src.management.commands.core import create_arc
from src.models import Arc, Vertex, Graph


def path(n, binary_number):
    with transaction.atomic():
        graph = create_path(n, binary_number)

        arc_count_in_subquery = (
            Arc.objects
            .filter(vertex_in=OuterRef('pk'))
            .values('vertex_in')
            .annotate(arc_count_in=Count('label'))
            .values('arc_count_in')
            .order_by('-arc_count_in')
            [:1]
        )

        arc_count_out_subquery = (
            Arc.objects
            .filter(vertex_out=OuterRef('pk'))
            .values('vertex_out')
            .annotate(arc_count_out=Count('label'))
            .values('arc_count_out')
            .order_by('-arc_count_out')
            [:1]
        )

        # Získání maximálního počtu arků pro každý vrchol
        vertices = (
            Vertex.objects
            .filter(graph=graph)
            .annotate(arc_count_in=Coalesce(Subquery(arc_count_in_subquery, output_field=IntegerField()), 0))
            .annotate(arc_count_out=Coalesce(Subquery(arc_count_out_subquery, output_field=IntegerField()), 0))
            .annotate(max_arc_count=ExpressionWrapper(
                Max(F('arc_count_in'), F('arc_count_out')),
                output_field=IntegerField()
            ))
            .order_by('max_arc_count', 'label')
        )
        for num, v in enumerate(vertices):
            v.label = num + 1
            v.save()
        graph.update()


def create_path(n, binary_number, graph=None):
    prev_vertex = None
    arc_num = 0
    if graph:
        graph.components += 1
        graph.graph_type = Graph.UNITED
        graph.save()
    else:
        graph = Graph.objects.create(
            graph_type=Graph.PATH,
            binary=binary_number[1:],
        )

    for j, bit in enumerate(binary_number):
        vertex = Vertex.objects.create(graph=graph, index=j + 1)
        if prev_vertex:
            arc_num = create_arc(
                vertex_in=prev_vertex if bit == '0' else vertex,
                vertex_out=vertex if bit == '0' else prev_vertex,
                arc_num=arc_num,
                label=n + arc_num
            )
        prev_vertex = vertex
    return graph