from django.db import models


class Graph(models.Model):
    CYCLE = "cycle"
    PATH = "path"
    TADPOLE = "tadpole"
    UNITED = "united"

    GRAPHS = (
        (CYCLE, "cycle"),
        (PATH, "path"),
        (TADPOLE, "tadpole"),
        (UNITED, "united")
    )

    graph_type = models.CharField(max_length=8, choices=GRAPHS)
    binary = models.CharField(max_length=2048, blank=True, null=True)
    components = models.IntegerField(default=1)

    antimagic = models.BooleanField(default=False)
    strong = models.BooleanField(default=False)

    @property
    def vertex_count(self):
        return self.vertices.filter(graph=self).count()

    @property
    def arc_count(self):
        return self.vertices.filter(in_arcs__isnull=False).count()
        # return Arc.objects.filter(vertex_in__graph=self).count()

    def get_properties(self):
        return {
            'vertex': self.vertex_count,
            'arc': self.arc_count,
        }

    def update(self):
        vertices = self.vertices.all()
        in_weight = []
        out_weight = []
        labels = []

        for vertex in vertices:
            labels.append(vertex.label)
            in_weight.append(vertex.in_weight)
            out_weight.append(vertex.out_weight)

        self.antimagic = len(in_weight) == len(set(in_weight)) and len(out_weight) == len(set(out_weight))
        self.strong = len(in_weight + out_weight) == len(set(in_weight + out_weight))
        self.save()

    def __str__(self):
        return f"{self.graph_type}, {self.vertex_count};{self.arc_count}"


class Vertex(models.Model):
    graph = models.ForeignKey(Graph, on_delete=models.CASCADE, blank=True, null=True, related_name="vertices")
    index = models.IntegerField(default=0, null=True, blank=True)
    label = models.IntegerField(default=0, null=True, blank=True)

    @property
    def max_degree(self):
        sum_in = self.in_arcs.all().count()
        sum_out = self.out_arcs.all().count()
        # in_arcs = Arc.objects.filter(vertex_in=self)
        # sum_in = sum(1 for _ in in_arcs)
        # out_arcs = Arc.objects.filter(vertex_out=self)
        # sum_out = sum(1 for _ in out_arcs)
        return sum_in if sum_in > sum_out else sum_out

    @property
    def in_weight(self):
        in_arcs = Arc.objects.filter(vertex_in=self)
        return self.label + sum(arc.label for arc in in_arcs)

    @property
    def out_weight(self):
        out_arcs = Arc.objects.filter(vertex_out=self)
        return self.label + sum(arc.label for arc in out_arcs)

    def get_properties(self):
        return {
            'in_weight': self.in_weight,
            'out_weight': self.out_weight,
            'max_degree': self.max_degree,
        }

    def __str__(self):
        return f"{self.id}, {self.label}, {self.index}"


class Arc(models.Model):
    vertex_in = models.ForeignKey(Vertex, on_delete=models.CASCADE, related_name='in_arcs')
    vertex_out = models.ForeignKey(Vertex, on_delete=models.CASCADE, related_name='out_arcs')
    label = models.IntegerField(default=0, null=True, blank=True)

    def __str__(self):
        return f"{self.id}, ({self.vertex_in.label}, {self.vertex_out.label}) {self.label}"
